import logging
import os
import shutil
from distutils.dir_util import copy_tree

import setuptools
from piper.cli.setupper import Setupper
from piper.tools.venv import Virtualenv

logger = logging.getLogger(__name__)


def pack(context: str, package: str, command: str):

    # Setup the pipes and dependencies
    current_pipe, all_pipes = Setupper().manage(context, package=package)

    # Get the right python (this might be useful in case the setup.py uses a particular syntax)
    env = Virtualenv(all_pipes[package])

    # Clean the previous target folder, if existing
    build_folder = current_pipe.build_folder
    if os.path.isdir(build_folder):
        shutil.rmtree(build_folder)

    # Copy the structure of this package
    _copy_folder(current_pipe.setup_py_folder, build_folder)
    source_build_folder = os.path.join(build_folder, os.path.basename(current_pipe.setup_py_folder))
    # For all dependencies, copy their code too
    for dep in current_pipe.flat_dependencies():
        # For all code packages, copy it inside the copied setup.py folder
        for folder in setuptools.find_packages(where=dep.setup_py_folder):
            _copy_folder(
                os.path.join(dep.setup_py_folder, folder),
                os.path.join(source_build_folder)
            )

    # Setup the code
    setup_py = os.path.join(source_build_folder, "setup.py")
    env.python.run(f"{setup_py} {' '.join(command)}", cwd=source_build_folder)  # TODO should check for spaces and commas, or use list!


def _copy_folder(source, destination):
    os.makedirs(destination, exist_ok=True)
    right_destination = os.path.join(destination, os.path.basename(os.path.normpath(source)))
    return copy_tree(source, right_destination)


import logging
import os
import shutil

from piper.cli.base import PiperCommand
from piper.config.pipe import Pipe
from piper.tools.venv import Virtualenv

logger = logging.getLogger(__name__)


class Cleaner(PiperCommand):

    def _manage(self, pipe: Pipe):

        logger.info(f"Cleaning {pipe.name}")
        to_remove = [
            os.path.join(pipe.setup_py_folder, f"{pipe.package}.egg-info"),     # egg-info
            Virtualenv(pipe).path,                                              # venv folder
            pipe.build_folder                                                   # artifacts build folder
        ]

        # Clean
        for path in to_remove:
            if os.path.isdir(path):
                shutil.rmtree(path)



import re
from typing import Optional

from piper.config.sanitizers import asserter

_VERSION = re.compile(r"^\d+\.\d+.\d+$")


def sanitize_version(version: str, nullable: bool = False) -> Optional[str]:
    if version is None and nullable:
        return None
    return asserter.regex(_VERSION, version)



import logging
import os
import shutil
from typing import Dict, Any

import pkg_resources
import yaml
from piper import RESOURCES_LOCATION
from piper.config.pipe import Pipe
from piper.config.sanitizers import pip_sanitizer, pipe_sanitizer
from piper.tools.venv import Virtualenv

logger = logging.getLogger(__name__)
_GENERATE_BLUEPRINT = os.path.join(RESOURCES_LOCATION, "blueprint", "generate.py")


def _generate_breath_first(blueprint_file: str, venv: Virtualenv, context: str, yml: Dict[str, Any]):

    # Generate these blueprints
    for pipe, properties in yml.items():

        # Get the folder where the pipe will be created
        pipe = pipe_sanitizer.sanitize_name(pipe)
        pipe_context = os.path.join(context, pipe)

        # Is it a blueprint?
        if "blueprint" in properties:

            # Get the package of the blueprint
            blueprint = properties["blueprint"]
            if len(blueprint) != 1:
                raise ValueError(f"Error in '{pipe}': this pipe's blueprint must be a single 'name: version_or_path' dictionary")
            name = pip_sanitizer.sanitize_package(list(blueprint.keys())[0])
            venv.pip.install(blueprint, allow_local_files=True)

            # Generate the blueprint
            command = f'"{_GENERATE_BLUEPRINT}" --context "{pipe_context}" --blueprint {name} --file "{blueprint_file}" --pipe {pipe}'
            venv.python.run(command)

    # Generate the subpipes
    for pipe, properties in yml.items():
        pipe_context = os.path.join(context, pipe)
        if "pipes" in properties:
            _generate_breath_first(blueprint_file, venv, pipe_context, properties["pipes"])


def run(context: str):

    # Temporary create a venv
    venv = Virtualenv(Pipe(location=context, yml={}))
    venv.create()
    version = pip_sanitizer.sanitize_version(pkg_resources.get_distribution("PyYAML").version)
    venv.pip.install({"PyYAML": version})

    # Read the blueprint file
    path = os.path.join(context, "blueprint.yml")
    with open(path) as f:
        yml = yaml.load(f, Loader=yaml.SafeLoader)
    if not yml or "pipes" not in yml:
        raise ValueError("Blueprint file must not be empty and should contain at least 'pipes' key.")

    # Generate
    _generate_breath_first(path, venv, context, yml["pipes"])

    # Remove the temporary venv
    shutil.rmtree(venv.path)



import logging
import sys

import os

PACKAGE_LOCATION = os.path.join(os.path.dirname(os.path.realpath(__file__)))
RESOURCES_LOCATION = os.path.join(PACKAGE_LOCATION, "resources")
CONFIG_LOCATION = os.path.join(RESOURCES_LOCATION, "config")


def _init_logger():
    root_log = logging.getLogger()
    formatter = logging.Formatter("%(message)s")
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setFormatter(formatter)
    root_log.addHandler(stream_handler)
    root_log.setLevel(logging.INFO)


_init_logger()



import os
import sys
from typing import List

from piper.tools.executable import Executable


class Python(Executable):

    def __init__(self, path: str = None, pre_command: str = None):
        self.path = path or os.environ.get("PIPER_PYTHON_PATH", sys.executable)
        super().__init__(self.path, pre_command=pre_command)
        # TODO version from Popen

    def replace_process(self, command: List[str]):
        os.execv(self.path, [os.path.basename(os.path.normpath(self.path)), *command])

main_python = Python()



import re

from piper.config.sanitizers import asserter

_PACKAGE = re.compile(r"^[A-Za-z0-9\-\._]+$")
_VERSION = re.compile(r"^[A-Za-z0-9\-\._]+$")  # To be changed


def sanitize_package(package: str) -> str:
    return asserter.regex(_PACKAGE, package)


def sanitize_version(version: str) -> str:
    return asserter.regex(_VERSION, version)



import os
from pathlib import Path

from piper.cli.base import PiperCommand
from piper.config.pipe import Pipe
from piper.tools.venv import Virtualenv


class Setupper(PiperCommand):

    def _manage(self, pipe: Pipe):

        # Prepare the requirements file
        if not os.path.isfile(pipe.requirements_file):
            Path(pipe.requirements_file).touch()

        # Create the virtualenv
        venv = Virtualenv(pipe)
        venv.create()
        venv.update()

        # Pip freeze
        venv.freeze()



import logging
import subprocess

from piper.tools.shell import Shell

logger = logging.getLogger(__name__)


class Executable(Shell):

    def __init__(self, tool: str, pre_command: str = None):
        super().__init__(pre_command)
        self._tool = tool

    def run(self, command: str, capture: bool = False, text: bool = True, cwd: str = None) -> subprocess.CompletedProcess:
        return super().run(f"{self._tool} {command}", capture=capture, text=text, cwd=cwd)



import logging
import os
from pathlib import Path
from typing import List, Dict

import yaml
from piper.config.pipe import Pipe

_PIPE_FILE = "pipe.yml"


logger = logging.getLogger(__name__)


def _read_pipe(location: str) -> Pipe:

    # Read the pipe file
    path = os.path.join(location, _PIPE_FILE)
    with open(path) as f:
        yml = yaml.load(f, Loader=yaml.SafeLoader) or {}
        return Pipe(location=location, yml=yml)


def _has_pipe(location: str) -> bool:
    return os.path.isfile(os.path.join(location, _PIPE_FILE))


def get_pipe_name(location: str) -> str:
    return os.path.basename(os.path.normpath(location))


def read_all_pipes(location: str) -> Dict[str, Pipe]:

    # Find the uppermost (parent) pipe file
    uppermost = location
    while _has_pipe(location):
        uppermost = location
        location = str(Path(location).parent)

    # Read main pipe file and nested pipe files
    pipes = {pipe.name: pipe for pipe in _read_pipes_from(uppermost)}

    # Build dependencies
    for name in pipes:
        for pipe in pipes.values():
            if name in pipe.dependencies:
                pipe.fill_dependency_with_pipe(pipes[name])
    return pipes


def _read_pipes_from(location: str) -> List[Pipe]:

    # Read the local pipe
    group = []
    pipe = _read_pipe(location)

    # Read all pipes recursively
    pipes = [pipe]
    for directory in os.listdir(location):
        current = os.path.join(location, directory)
        if _has_pipe(current):
            children = _read_pipes_from(current)
            group = group + children
            pipes = pipes + children

    # Set the nested group
    pipe.group = group
    return pipes



import abc
import logging
from typing import Set

from piper.config.pipe import Pipe
from piper.config.reader import get_pipe_name, read_all_pipes
from piper.tools.python import main_python, Python


logger = logging.getLogger(__name__)


class PiperCommand(abc.ABC):

    def __init__(self, python: Python = main_python):
        self.main_python = python

    @abc.abstractmethod
    def _manage(self, pipe: Pipe):
        pass

    def _manage_from_pipe_recursive(self, pipe: Pipe, done: Set[Pipe]):

        if not pipe.group:

            # Manage dependencies first, then the pipe
            for p in [*pipe.flat_dependencies(), pipe]:
                if p not in done:
                    logger.info("")
                    logger.info("="*60)
                    logger.info(f"=== {p.name.center(52, ' ')} ===")
                    logger.info("="*60)
                    self._manage(p)
                    logger.info("=" * 60)
                    logger.info("")
                    logger.info("")
                    done.add(p)

        else:

            # Manage each element of the group
            for pipe in pipe.group:
                self._manage_from_pipe_recursive(pipe, done=done)

    def manage_from_pipe(self, pipe: Pipe):
        self._manage_from_pipe_recursive(pipe, done=set())

    def manage(self, context: str, package: str = None) -> (Pipe, Pipe):

        # Read the target pipe and all the pipes
        target = package or get_pipe_name(context)
        pipes = read_all_pipes(context)

        # Manage this target pipe
        if target not in pipes:
            raise ValueError(f"The specified pipe does not exist: {context}")
        pipe = pipes[target]
        self.manage_from_pipe(pipe)
        return pipe, pipes



import os
from typing import Dict

from piper.tools.executable import Executable
from piper.config.sanitizers.pip_sanitizer import sanitize_package, sanitize_version

class Pip(Executable):

    def __init__(self, pre_command: str = None):
        super().__init__("pip", pre_command=pre_command)

    def install(self, requirements: Dict[str, str], allow_local_files=False):
        packages = [
            version if allow_local_files and os.path.isdir(version) else f'{sanitize_package(name)}=={sanitize_version(version)}'
            for name, version in requirements.items()
        ]
        super().run(f"install {' '.join(packages)}")



import logging
import os
import re
import subprocess

from piper.config.pipe import Pipe
from piper.tools.pip import Pip
from piper.tools.python import Python
from piper.tools.shell import Shell
from piper.tools.python import main_python as default_python

logger = logging.getLogger(__name__)


_CREATE_VENV = "(" + ") || (".join(["{python} -m venv {venv}",
                                    "virtualenv -p {python} {venv}",
                                    "pip install virtualenv && virtualenv -p {python} {venv}"]) + ")"


class Virtualenv:

    def __init__(self, pipe: Pipe, main_python: Python = default_python):
        self._main_python = main_python
        self._name = f"venv-{pipe.name}"

        self.pipe = pipe
        self.path = os.path.join(pipe.location, self._name)
        activate = _get_activate_command(self.path)
        self.python = Python(path=os.path.join(self.path, "bin" if os.name == "posix" else "Scripts", "python" if os.name == "posix" else "python.exe"), pre_command=activate)
        self.pip = Pip(pre_command=activate)

    def create(self) -> None:
        # Create the virtualenv if it does not exist
        if os.path.exists(self.path):
            logger.info(f"Virtualenv already exists for {self._name}")
        else:
            # Create the virtualenv with the template command
            logger.info(f"Creating the virtual environment for {self._name}...")
            Shell().run(
                _CREATE_VENV.format(
                    python=self._main_python.path,
                    venv=self.path
                )
            )

    def update(self):

        # Install requirements
        logger.info("Installing requirements...")
        for pipe in [*self.pipe.flat_dependencies(), self.pipe]:
            self.pip.run(f"install -e {pipe.setup_py_folder}")
            if pipe.requirements:
                self.pip.run(f"install {' '.join([f'{name}=={version}' for name, version in pipe.requirements.items()])}")
        logger.info("Done.")

    def freeze(self):

        try:

            # Pip freeze
            output = self.pip.run(f"freeze", capture=True).stdout
            requirements = re.sub("[\r\n]+", "\n", output).split("\n")

            # Get real requirements and update the requirements.txt file
            real_requirements = [req for req in requirements if _is_valid_requirement(req)]
            with open(self.pipe.requirements_file, "w") as f:
                for requirement in real_requirements:
                    f.write(requirement + "\n")
                logger.info(f"Freezing {real_requirements}")

        except subprocess.CalledProcessError as e:
            logger.error(e.stderr)
            exit(e.returncode)


def _is_valid_requirement(requirement: str) -> bool:

    # List of requirements to filter out
    forbidden_requirements = [
        "pkg-resources",
        "^-e",
        "^#",
        "pywin32",
        "pywinpty"
    ]
    forbidden_match = "(" + ")|(".join(forbidden_requirements) + ")"

    # Check if there is one match
    return requirement and not re.match(forbidden_match, requirement)


def _get_activate_command(path) -> str:
    # Get activate command
    if os.name == "nt":
        activate = f"{path}\\Scripts\\activate"
    elif os.name == "posix":
        activate = f". {path}/bin/activate"
    else:
        raise NotImplementedError(f"Operating System {os.name} not supported.")
    return activate



import argparse
import importlib
import inspect
import sys
from pathlib import Path

import yaml
from piperblue.blueprint import PiperBlueprint


def _find(yml, pipe):
    try:
        for p, properties in yml.items():
            if p == pipe:
                return properties
            found = _find(yml[p], pipe)
            if found:
                return found
    except:
        return None


def _generate(file: str, blueprint_name: str, pipe: str, context: str):

    # Find the config
    with open(file) as f:
        yml = yaml.load(f, Loader=yaml.SafeLoader)
    properties = _find(yml, pipe)
    config = properties.get("config")

    # Import the blueprint class
    module = f"{blueprint_name}.blueprint"
    all_blueprints = inspect.getmembers(
        importlib.import_module(module),
        lambda member: inspect.isclass(member) and member.__module__ == module and issubclass(member, PiperBlueprint)
    )[0]
    blueprint = all_blueprints[1]

    # Get its resources folder
    resources_folder = str(Path(sys.modules[blueprint.__module__].__file__).parent / "resources")

    # Instantiate it and run it
    blueprint().generate(resources_folder, context, config)


def main():

    # Parse
    parser = argparse.ArgumentParser()
    parser.add_argument("--file")
    parser.add_argument("--context")
    parser.add_argument("--pipe")
    parser.add_argument("--blueprint")
    args = parser.parse_args()

    # Generate
    _generate(args.file, args.blueprint, args.pipe, args.context)


if __name__ == '__main__':

    main()



import argparse
import logging
import os

from piper.cli import python_wrapper, blueprinter, packer
from piper.cli.cleaner import Cleaner
from piper.cli.setupper import Setupper

logger = logging.getLogger(__name__)


def main():

    # Piper
    parser = argparse.ArgumentParser()
    parser.add_argument("--context", default=os.getcwd(), help="Path to the folder containing the pipe.yml file")
    parser.set_defaults(action=lambda args: parser.parse_args(["-h"]))
    subparsers = parser.add_subparsers()

    # Clean
    clean = subparsers.add_parser("clean")
    clean.set_defaults(action=lambda args, rest: Cleaner().manage(args.context))

    # Blueprint
    blueprint = subparsers.add_parser("blueprint")
    blueprint.set_defaults(action=lambda args, rest: blueprinter.run(args.context))

    # Setup
    setup = subparsers.add_parser("setup")
    setup.set_defaults(action=lambda args, rest: Setupper().manage(args.context))

    # Pack
    pack = subparsers.add_parser("pack")
    pack.add_argument("package", help="Name of the pipe to run setup.py commands")
    pack.set_defaults(action=lambda args, rest: packer.pack(args.context, args.package, rest))

    # Release

    # # Python
    # python = subparsers.add_parser("python")
    # python.add_argument("package", help="Name of the pipe (it will be used to activate its venv)")
    # python.set_defaults(action=lambda args, rest: python_wrapper.run(args.context, args.package, rest))

    # Parse args and invoke the corresponding functions
    arguments, remainder = parser.parse_known_args()
    arguments.action(arguments, remainder)
    logger.info("All done.")


if __name__ == '__main__':
    main()



import logging
import queue
import subprocess
from queue import Queue
from threading import Thread
from typing import Optional

logger = logging.getLogger(__name__)

_STDOUT = "stdout"
_STDERR = "stderr"


class Shell:

    def __init__(self, pre_command: str = None):
        self._pre_command = f"{pre_command} && " if pre_command else ""

    def run(self, command: str, capture: bool = False, text: bool = True, cwd: str = None) -> Optional[subprocess.CompletedProcess]:
        if capture:
            return self.captured_run(command, text=text, cwd=cwd)
        else:
            self.live_run(command, text=text, cwd=cwd)

    def captured_run(self, command: str, text: bool = True, cwd: str = None):
        return subprocess.run(
            self._pre_command + command,
            check=True,
            capture_output=True,
            text=text,
            shell=True,
            cwd=cwd
        )

    def live_run(self, command: str, text: bool = True, cwd: str = None):

        output = Queue()

        def pipe(std, stream):
            # Blocking for loop: the stream will be empty only on process termination
            for line in stream:
                output.put((std, line))
            if not stream.closed:
                stream.close()

        # Instantiate the child process
        process = subprocess.Popen(
            self._pre_command + command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=text,
            shell=True,
            cwd=cwd
        )

        # Start the producers
        stdout = Thread(target=pipe, args=(_STDOUT, process.stdout))
        stdout.start()
        stderr = Thread(target=pipe, args=(_STDERR, process.stderr))
        stderr.start()

        # Be the consumer
        outcome = None
        while outcome is None:

            try:

                try:
                    # Get a line, waiting at most one second (cannot be less)
                    line = output.get(timeout=0.2)
                except queue.Empty:
                    # Nothing to do
                    pass
                else:
                    # Print that line
                    std, msg = line
                    msg = msg.rstrip("\r\n") if text else msg
                    if std == _STDOUT:
                        logger.info(msg)
                    else:
                        logger.error(msg)
                finally:
                    # Check if the process has ended
                    outcome = process.poll()

            except KeyboardInterrupt:

                process.kill()
                outcome = -9

        # Release resources (if possible)
        stdout.join(1)
        stderr.join(1)
        if outcome != 0:
            exit(outcome)
        return outcome



from setuptools import setup, find_packages
import os


# Function to retrieve resources files
def _get_resources(package_name):
    # Get all the resources (also on nested levels)
    res_paths = os.path.join(package_name, "resources")
    all_resources = [os.path.join(folder, file) for folder, _, files in os.walk(res_paths) for file in files]
    # Remove the prefix: start just from "resources"
    return [resource[resource.index("resources"):] for resource in all_resources]


# Read requirements
this_location = os.path.dirname(os.path.realpath(__file__))
requirements_file = os.path.join(this_location, "requirements.txt")
with open(requirements_file, "r") as f:
    requirements = f.read().splitlines()

# Read README
readme_file = os.path.join(this_location, "README.md")
with open(readme_file, "r", encoding="utf-8") as f:
    readme = f.read()


# Package configuration
_NAME="piper-tools"
setup(
    name=_NAME,
    version="0.1.0",
    description="A software manager for easy development and distribution of Python code",
    packages=find_packages(),
    package_data={"piper": _get_resources("piper")},
    install_requires=requirements,
    python_requires=">=3.7",
    entry_points={
        'console_scripts': [
            'piper = piper.__main__:main',
        ],
    },

    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/piper-tools/piper",
    author="Federico Pugliese",
    author_email="federico.pugliese.wr@gmail.com",
    license="Apache Software License, Version 2.0",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7"
    ]
)



import logging
import os

from piper.config.sanitizers import python_sanitizer, pip_sanitizer

logger = logging.getLogger(__name__)


class Pipe:

    def __init__(self, location: str, yml: dict):
        self.location = location
        self.name = os.path.basename(os.path.normpath(location))
        self.group = []

        # Pop the config entries
        config = yml.copy()
        self.python = python_sanitizer.sanitize_version(config.pop("python", None), nullable=True)
        self.dependencies = {pip_sanitizer.sanitize_package(dep): None for dep in config.pop("dependencies", [])}
        self.requirements = {pip_sanitizer.sanitize_package(name): pip_sanitizer.sanitize_version(version) for name, version in config.pop("requirements", {}).items()}

        # Extra utils
        self.package = self.name  # But could be different
        self.setup_py_folder = os.path.join(self.location, "package")  # But could be different
        self.build_folder = os.path.join(self.location, "build")  # But could be different
        self.requirements_file = os.path.join(self.setup_py_folder, "requirements.txt")

        # If there are still configs, they are unknown. Raise an error
        if config:
            raise KeyError(f"Unknown configuration in pipe file: {config}")

    def fill_dependency_with_pipe(self, pipe):
        self.dependencies[pipe.name] = pipe

    def flat_dependencies(self, with_current: bool = False):

        def visit(pipe, visited, level=0):
            for dep in pipe.dependencies.values():
                visited[dep] = max(level, visited.get(dep, level))
                visit(dep, visited, level=level+1)
            return visited

        pipes = visit(self, {})
        dependencies = [pipe for pipe, level in sorted(pipes.items(), key=lambda x: x[1], reverse=True)]
        if with_current:
            dependencies.append(self)
        return dependencies

    def __str__(self):
        return str(vars(self))

    def __repr__(self):
        return str(self)



import logging
import os

from piper.cli.setupper import Setupper
from piper.tools.python import main_python
from piper.tools.venv import Virtualenv

logger = logging.getLogger(__name__)


def run(context: str, package: str, command: str):

    # Setup the pipes and dependencies
    current_pipe, all_pipes = Setupper().manage(context, package=package)

    # Run the code
    env = Virtualenv(all_pipes[package])
    if os.path.isdir(env.path):
        python = env.python
    else:
        python = main_python

    # Good in unix
    if os.name == "posix":
        python.replace_process(command)
    else:
        logger.error(f" !!! Run the same command with {python.path}")
